﻿using SenseHat.Helpers;
using SenseHat.Sensors;
using SenseHatIO.SenseHatLedArray;
using System;
using System.Threading.Tasks;
using Windows.UI;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Navigation;

namespace SpaceDot
{
    public sealed partial class MainPage : Page
    {
        private InertialSensor inertialSensor = InertialSensor.Instance;
        private LedArray ledArray;

        // Initial position, and default color
        private int x = 3;
        private int y = 3;
        private Color dotColor = Colors.Blue;

        private const float accelerationScaler = 2.5f;

        public MainPage()
        {
            InitializeComponent();
        }

        protected async override void OnNavigatedTo(NavigationEventArgs e)
        {
            base.OnNavigatedTo(e);

            // Initialize led array
            await InitLedArray();

            // Begin sensor readings
            if (await inertialSensor.Initialize())
            {
                BeginAccelerationAcquisition();
            }
        }

        private async Task InitLedArray()
        {
            const byte address = 0x46;
            var device = await I2cHelper.GetI2cDevice(address);

            if (device != null)
            {
                ledArray = new LedArray(device);
                ledArray.SetPixel(x, y, dotColor);
            }
            else
            {
                throw new Exception("Device could not be initialized");
            }
        }

        private void BeginAccelerationAcquisition()
        {
            const int msDelayTime = 25;

            BeginSensorReading(() =>
            {
                var linearAcceleration = inertialSensor.GetLinearAcceleration();

                UpdateDotPosition(linearAcceleration);
            }, msDelayTime);
        }

        private void UpdateDotPosition(Vector3D<float> accelerometerReading)
        {
            var stepX = Convert.ToInt32(accelerometerReading.X * accelerationScaler);
            var stepY = Convert.ToInt32(accelerometerReading.Y * accelerationScaler);

            x = CorrectLedCoordinate(x - stepX);
            y = CorrectLedCoordinate(y - stepY);

            ledArray.SetPixel(x, y, dotColor);
        }

        private static int CorrectLedCoordinate(int inputCoordinate)
        {
            inputCoordinate = Math.Min(inputCoordinate, LedArray.Length - 1);
            inputCoordinate = Math.Max(inputCoordinate, 0);

            return inputCoordinate;
        }

        private void BeginSensorReading(Action periodicAction, int msDelayTime)
        {
            Task.Run(() =>
            {
                while (true)
                {
                    periodicAction();

                    Task.Delay(msDelayTime).Wait();
                }
            });
        }
    }
}
