﻿using System;
using Windows.Media.SpeechRecognition;
using Windows.Media.SpeechSynthesis;
using Windows.UI.Core;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;

namespace Speech
{
    public sealed partial class MainPage : Page
    {
        private string textToSpeak;

        private object allVoices = SpeechSynthesizer.AllVoices;
        private object voice;

        private SpeechRecognizer speechRecognizer;

        private const string startCaption = "Start";
        private const string stopCaption = "Stop";

        public MainPage()
        {
            InitializeComponent();

            InitializeSpeechRecognizer();            
        }

        private void ButtonSpeak_Click(object sender, RoutedEventArgs e)
        {
            SpeechHelper.Speak(textToSpeak, voice as VoiceInformation);
        }

        private async void InitializeSpeechRecognizer()
        {
            speechRecognizer = new SpeechRecognizer();            
            
            await speechRecognizer.CompileConstraintsAsync();

            speechRecognizer.RecognitionQualityDegrading += SpeechRecognizer_RecognitionQualityDegrading;
            speechRecognizer.StateChanged += SpeechRecognizer_StateChanged;
            speechRecognizer.HypothesisGenerated += SpeechRecognizer_HypothesisGenerated;                        

            speechRecognizer.ContinuousRecognitionSession.ResultGenerated += ContinuousRecognitionSession_ResultGenerated;
            speechRecognizer.ContinuousRecognitionSession.Completed += ContinuousRecognitionSession_Completed;            
        }

        private void SpeechRecognizer_RecognitionQualityDegrading(SpeechRecognizer sender, SpeechRecognitionQualityDegradingEventArgs args)
        {
            DisplayInfo("Quality degrading: " + args.Problem);
        }

        private void SpeechRecognizer_StateChanged(SpeechRecognizer sender, SpeechRecognizerStateChangedEventArgs args)
        {
            DisplayInfo("State changed: " + args.State);
        }

        private void SpeechRecognizer_HypothesisGenerated(SpeechRecognizer sender, SpeechRecognitionHypothesisGeneratedEventArgs args)
        {
            DisplayInfo("Hypothesis: " + args.Hypothesis.Text);
        }        

        private async void ButtonOneTimeRecognition_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                var recognitionResult = await speechRecognizer.RecognizeAsync();

                DisplayInfo(GetRecognitionResultInfo(recognitionResult));
            }
            catch (UnauthorizedAccessException)
            {
                DisplayInfo("Speech recognition requires an access to a microphone");
            }
            catch (Exception)
            {
                DisplayInfo("Speech recognition is disabled");
            }
        }

        private async void DisplayInfo(string infoMessage)
        {
            if(Dispatcher.HasThreadAccess)
            {
                ListBoxResults.Items.Add(infoMessage);
                ListBoxResults.SelectedIndex = ListBoxResults.Items.Count - 1;                
            }
            else
            {
                await Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () => {
                    DisplayInfo(infoMessage);
                });
            }
        }        

        private string GetRecognitionResultInfo(SpeechRecognitionResult speechRecognitionResult)
        {
            return string.Format("Recognition result. Status: {0}. Confidence: {1}. Text: {2}",
                speechRecognitionResult.Status,
                speechRecognitionResult.Confidence,
                speechRecognitionResult.Text);
        }

        private void ButtonClearList_Click(object sender, RoutedEventArgs e)
        {
            ListBoxResults.Items.Clear();
        }

        private async void ButtonContinousRecognition_Click(object sender, RoutedEventArgs e)
        {
            string buttonCaption = ButtonContinousRecognition.Content.ToString();

            if (buttonCaption.Contains(startCaption))
            {
                try
                {
                    await speechRecognizer.ContinuousRecognitionSession.StartAsync();

                    ButtonContinousRecognition.Content = buttonCaption.Replace(startCaption, stopCaption);
                }
                catch (UnauthorizedAccessException)
                {
                    DisplayInfo("Speech recognition requires an access to a microphone");
                }
                catch (Exception)
                {
                    DisplayInfo("Speech recognition is disabled");
                }
            }
            else
            {
                await speechRecognizer.ContinuousRecognitionSession.StopAsync();

                ButtonContinousRecognition.Content = buttonCaption.Replace(stopCaption, startCaption);
            }
        }

        private void ContinuousRecognitionSession_ResultGenerated(SpeechContinuousRecognitionSession sender,
            SpeechContinuousRecognitionResultGeneratedEventArgs args)
        {
            DisplayInfo(GetRecognitionResultInfo(args.Result));            
        }

        private void ContinuousRecognitionSession_Completed(SpeechContinuousRecognitionSession sender, SpeechContinuousRecognitionCompletedEventArgs args)
        {
            DisplayInfo("Speech recognition completed. Status: " + args.Status);
        }
    }
}
