﻿using AnomalyDetection.PredictiveClient;
using AnomalyDetection.Training;
using AnomalyDetection.ViewModels;
using SenseHat.Helpers;
using SenseHat.Sensors;
using SenseHatIO.SenseHatLedArray;
using System;
using System.Diagnostics;
using System.Threading;
using System.Threading.Tasks;
using Windows.UI;
using Windows.UI.Core;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;

namespace AnomalyDetection
{
    public sealed partial class MainPage : Page
    {
        private bool isSensorEmulationMode = true;
        private bool isLedArrayAvailable = false;

        private AnomalyViewModel anomalyViewModel;

        private SolidColorBrush normalTemperatureLevelColorBrush = new SolidColorBrush(Colors.Green);
        private SolidColorBrush abnormalTemperatureLevelColorBrush = new SolidColorBrush(Colors.Red);

        private TrainingDatasetAcquisition trainingDatasetAcquisition = new TrainingDatasetAcquisition();

        private PredictiveServiceClient predictiveServiceClient = new PredictiveServiceClient();

        private TemperatureAndPressureSensor sensor = TemperatureAndPressureSensor.Instance;

        private Task temperatureMonitoringTask;
        private CancellationTokenSource temperatureMonitoringCancellationTokenSource;

        private LedArray ledArray;

        public MainPage()
        {
            InitializeComponent();

            anomalyViewModel = new AnomalyViewModel()
            {
                IsAcquisitionInProgress = false,
                IsTemperatureMonitoringEnabled = false
            };
        }

        protected override async void OnNavigatedTo(NavigationEventArgs e)
        {
            base.OnNavigatedTo(e);

            await sensor.Initialize();

            await InitializeLedArray();
        }

        private async Task InitializeLedArray()
        {         
            const byte address = 0x46;
            var device = await I2cHelper.GetI2cDevice(address);

            if (device != null)
            {
                ledArray = new LedArray(device);
                ledArray.Reset(Colors.Black);

                isLedArrayAvailable = true;
            }
        }

        private async void ButtonAcquireTrainingDataset_Click(object sender, RoutedEventArgs e)
        {
            const int msDelay = 40;
            const int secDuration = 30;

            anomalyViewModel.IsAcquisitionInProgress = true;

            anomalyViewModel.FilePath = await trainingDatasetAcquisition.Acquire(msDelay, TimeSpan.FromSeconds(secDuration));

            anomalyViewModel.IsAcquisitionInProgress = false;
        }

        private async void GetAndProcessTemperature()
        {
            try
            {
                var temperature = GetTemperature();

                var result = await predictiveServiceClient.PredictAnomalyAsync(temperature);

                UpdateTemperatureDisplay(temperature, result.IsNormal);                        
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
            }
        }

        private float GetTemperature()
        {
            if (isSensorEmulationMode)
            {
                return GetRandomTemperature();
            }
            else
            {
                return sensor.GetTemperature();
            }
        }

        private float GetRandomTemperature(float baseTemperature = 36.55f)
        {
            var random = new Random();

            const double scaler = 0.5;

            return (float)(baseTemperature + random.NextDouble() * scaler);
        }

        private void ButtonStartTemperatureMonitoring_Click(object sender, RoutedEventArgs e)
        {
            anomalyViewModel.IsTemperatureMonitoringEnabled = true;

            temperatureMonitoringCancellationTokenSource = new CancellationTokenSource();
            
            temperatureMonitoringTask = new Task(() =>
            {
                const int msDelay = 500;

                while (!temperatureMonitoringCancellationTokenSource.IsCancellationRequested)
                {
                    GetAndProcessTemperature();

                    Task.Delay(msDelay).Wait();
                }

            }, temperatureMonitoringCancellationTokenSource.Token);

            temperatureMonitoringTask.Start();
        }

        private void ButtonStopTemperatureMonitoring_Click(object sender, RoutedEventArgs e)
        {
            temperatureMonitoringCancellationTokenSource.Cancel();
            temperatureMonitoringTask.Wait();

            anomalyViewModel.IsTemperatureMonitoringEnabled = false;
        }
        
        private async void UpdateTemperatureDisplay(float temperature, bool isNormalLevel)
        {
            if (Dispatcher.HasThreadAccess)
            {
                anomalyViewModel.Temperature = temperature;
                anomalyViewModel.TemperatureStatusColorBrush =
                    isNormalLevel ? normalTemperatureLevelColorBrush : abnormalTemperatureLevelColorBrush;

                UpdateLedArray();
            }
            else
            {
                await Dispatcher.RunAsync(CoreDispatcherPriority.Normal,
                    () => { UpdateTemperatureDisplay(temperature, isNormalLevel); });
            }
        }

        private void UpdateLedArray()
        {
            if (isLedArrayAvailable)
            {
                ledArray.Reset(anomalyViewModel.TemperatureStatusColorBrush.Color);
            }
        }
    }
}
