﻿using SenseHat.Portable.Helpers;
using SenseHat.UWP.Sensors;
using System;
using System.Threading;
using System.Threading.Tasks;

namespace CustomIoTSolution.RemoteClient.TelemetryControl
{
    public class Telemetry
    {
        // Event
        public event EventHandler<TelemetryEventArgs> DataReady = delegate { };

        // Telemetry task
        public bool IsActive { get; private set; } = false;

        private Task telemetryTask;
        private CancellationTokenSource telemetryCancellationTokenSource;

        // Sensors
        private TemperatureAndPressureSensor temperatureAndPressureSensor = TemperatureAndPressureSensor.Instance;
        private HumidityAndTemperatureSensor humidityAndTemperatureSensor = HumidityAndTemperatureSensor.Instance;
        
        private TimeSpan readoutDelay;

        private bool isEmulationMode = true;

        public void Start()
        {
            if (!IsActive)
            {
                InitializeTelemetryTask();

                telemetryTask.Start();

                IsActive = true;
            }
        }

        public void Stop()
        {
            if (IsActive)
            {
                telemetryCancellationTokenSource.Cancel();

                IsActive = false;
            }
        }

        public static async Task<Telemetry> CreateAsync(TimeSpan readoutDelay)
        {
            Check.IsNull(readoutDelay);

            var telemetry = new Telemetry(readoutDelay);

            await telemetry.InitializeSensors();

            return telemetry;
        }

        private Telemetry(TimeSpan readoutDelay)
        {
            this.readoutDelay = readoutDelay;
        }

        private async Task InitializeSensors()
        {
            if (!isEmulationMode)
            {
                await temperatureAndPressureSensor.Initialize();
                VerifyInitialization(temperatureAndPressureSensor, "Temperature and pressure sensor is unavailable");

                await humidityAndTemperatureSensor.Initialize();
                VerifyInitialization(humidityAndTemperatureSensor, "Humidity sensor is unavailable");
            }
        }

        private void VerifyInitialization(SensorBase sensorBase, string exceptionMessage)
        {
            if (!sensorBase.IsInitialized)
            {
                throw new Exception(exceptionMessage);
            }
        }

        private void InitializeTelemetryTask()
        {
            telemetryCancellationTokenSource = new CancellationTokenSource();

            telemetryTask = new Task(() =>
            {
                while (!telemetryCancellationTokenSource.IsCancellationRequested)
                {
                    if (IsActive)
                    {
                        var telemetryEventArgs = GetSensorReadings();

                        DataReady(this, telemetryEventArgs);

                        Task.Delay(readoutDelay).Wait();
                    }
                }
            }, telemetryCancellationTokenSource.Token);
        }

        private TelemetryEventArgs GetSensorReadings()
        {
            if (!isEmulationMode)
            {
                var temperature = temperatureAndPressureSensor.GetTemperature();
                var humidity = humidityAndTemperatureSensor.GetHumidity();
                var pressure = temperatureAndPressureSensor.GetPressure();

                return new TelemetryEventArgs(temperature, humidity, pressure);
            }
            else
            {
                var random = new Random();

                var temperature = (float)(30.0 + random.NextDouble() * 5.0);
                var humidity = (float)(20.0 + random.NextDouble() * 10.0);
                var pressure = (float)(1005.0 + random.NextDouble() * 2.5);

                return new TelemetryEventArgs(temperature, humidity, pressure);
            }
        }
    }
}
